extern crate mediawiki;

use crate::entity_diff::EntityDiff;
use crate::*;
use std::collections::HashMap;
use std::sync::mpsc;
use std::thread;

const LOAD_ENTITIES_MAX_THREADS: usize = 10;

/// A container of `Entity` values.
/// This can load and cache entities individually, or in large batches.
/// It will load entities only once, and load groups of 50/500 entities in parallel.
#[derive(Debug, Default, Clone)]
pub struct EntityContainer {
    entities: HashMap<String, Entity>,
}

impl EntityContainer {
    /// Generates a new, empty `EntityContainer`
    pub fn new() -> EntityContainer {
        EntityContainer {
            entities: HashMap::<String, Entity>::new(),
        }
    }

    /// Loads (new) entities from the MediaWiki API
    pub fn load_entities(
        &mut self,
        api: &mediawiki::api::Api,
        entity_ids: &Vec<String>,
    ) -> Result<(), Box<::std::error::Error>> {
        //let entities_before = self.entities.len();
        let chunk_size = match api.user().is_bot() {
            true => 200, // Could be 500 but timeouts happen
            false => 50,
        };
        let to_load = entity_ids
            .iter()
            .filter(|entity_id| !entity_id.is_empty())
            .filter(|entity_id| !self.entities.contains_key(*entity_id))
            .map(|entity_id| entity_id.to_owned())
            .collect::<Vec<String>>();

        let (tx, rx) = mpsc::channel();
        let mut thread_list = vec![];
        let mut chunks: u64 = 0;
        for chunk in to_load.chunks(chunk_size) {
            chunks = chunks + 1;
            let ids = chunk.join("|");
            let params: HashMap<_, _> = vec![
                ("action", "wbgetentities"),
                ("ids", &ids),
                ("format", "json"),
            ]
            .into_iter()
            .map(|(k, v)| (k.to_string(), v.to_string()))
            .collect();
            let req = api.get_api_request_builder(&params, "GET")?;

            let tx = mpsc::Sender::clone(&tx);
            thread_list.push((req, tx));
        }

        for _num in 0..LOAD_ENTITIES_MAX_THREADS {
            match thread_list.pop() {
                Some(x) => {
                    thread::spawn(move || {
                        let response = x.0.send().expect("Getting response from API failed");
                        x.1.send(response).expect("Sending of result failed");
                    });
                }
                None => break,
            }
        }

        for _ in 0..chunks {
            let mut response = rx.recv()?;

            // Start next thread
            match thread_list.pop() {
                Some(x) => {
                    thread::spawn(move || {
                        let response = x.0.send().expect("Getting response from API failed");
                        x.1.send(response).expect("Sending of result failed");
                    });
                }
                None => {}
            }

            let j: serde_json::Value = response.json()?;
            let entities = match j["entities"].as_object() {
                Some(e) => e,
                None => continue,
            };
            for (_entity_id, entity_json) in entities {
                self.set_entity_from_json(entity_json)?;
            }
        }
        //println!("Loaded {} entities", self.entities.len() - entities_before);
        Ok(())
    }

    /// Adds an `entity` to the cache, based on its JSON representation
    pub fn set_entity_from_json(
        &mut self,
        entity_json: &serde_json::Value,
    ) -> Result<(), Box<::std::error::Error>> {
        let entity_id = match &entity_json["id"] {
            serde_json::Value::String(s) => s.to_string(),
            _ => {
                return Err(From::from(format!(
                    "Entity has no 'id' (string) field:{}",
                    &entity_json
                )));
            }
        };
        let entity = from_json::entity_from_json(&entity_json)?;
        self.entities.insert(entity_id.to_string(), entity);
        Ok(())
    }

    /// Loads a single entity. Returns the `Entity`, or an error
    pub fn load_entity<S: Into<String>>(
        &mut self,
        api: &mediawiki::api::Api,
        entity: S,
    ) -> Result<&Entity, Box<::std::error::Error>> {
        let entity: String = entity.into();
        self.load_entities(api, &vec![entity.clone()])?;
        match self.get_entity(entity.as_str()) {
            Some(e) => Ok(e),
            None => Err(From::from(format!("No such entity '{}'", &entity))),
        }
    }

    /// Returns `Some(entity)` with that ID from the cache, or `None`.
    /// This will _not_ load entities via the API! Use `load_entity` for that
    pub fn get_entity<S: Into<String>>(&self, entity_id: S) -> Option<&Entity> {
        let entity_id: String = entity_id.into();
        self.entities.get(&entity_id)
    }

    /// Checks if an entity is in the cache.
    /// Returns true or false.
    pub fn has_entity<S: Into<String>>(&self, entity_id: S) -> bool {
        let entity_id: String = entity_id.into();
        match self.entities.get(&entity_id) {
            Some(_) => true,
            None => false,
        }
    }

    /// Removes the entity with the given key from the cache, and returns `Some(entity)` or `None`
    pub fn remove_entity<S: Into<String>>(&mut self, entity_id: S) -> Option<Entity> {
        let entity_id: String = entity_id.into();
        self.entities.remove(&entity_id)
    }

    /// Removes the entities with the given keys from the cache
    pub fn remove_entities(&mut self, entity_ids: &Vec<String>) {
        for entity_id in entity_ids {
            self.remove_entity(entity_id.to_string());
        }
    }

    /// Removes the entities with the given keys from the cache, then reloads them from the API
    pub fn reload_entities(
        &mut self,
        api: &mediawiki::api::Api,
        entity_ids: &Vec<String>,
    ) -> Result<(), Box<::std::error::Error>> {
        self.remove_entities(entity_ids);
        self.load_entities(api, entity_ids)?;
        Ok(())
    }

    /// Returns the number of cached entities
    pub fn len(&self) -> usize {
        self.entities.len()
    }

    /// Clears the cache
    pub fn clear(&mut self) {
        self.entities.clear();
    }

    /// Applies a diff, updates the cache if possible, and returns the entity ID
    pub fn apply_diff(
        &mut self,
        api: &mut mediawiki::api::Api,
        diff: &EntityDiff,
    ) -> Option<String> {
        if diff.is_empty() {
            return None; // Nothing done
        }
        match diff.apply_diff(api, &diff) {
            Ok(json) => match EntityDiff::get_entity_id(&json) {
                Some(q) => match self.set_entity_from_json(&json) {
                    Ok(_) => Some(q),
                    _ => None,
                },
                None => None,
            },
            _ => None,
        }
    }
}

#[cfg(test)]
mod tests {
    use super::mediawiki::api::Api;
    use super::EntityContainer;

    #[test]
    fn entities_misc() {
        let api = Api::new("https://www.wikidata.org/w/api.php").unwrap();
        let mut ec = EntityContainer::new();
        assert_eq!(ec.len(), 0);
        ec.load_entities(&api, &vec!["Q42".to_string(), "Q12345".to_string()])
            .unwrap();
        /*
                assert_eq!(ec.len(), 2);
                ec.reload_entities(&api, &vec!["Q12345".to_string()])
                    .unwrap();
                assert!(ec.has_entity("Q12345"));
                assert_eq!(ec.len(), 2);
        */
        /*
                let q42 = ec.get_entity("Q42");
                assert_eq!(q42.unwrap().id(), "Q42");
                ec.remove_entity("Q42");
                assert_eq!(ec.len(), 1);
                ec.clear();
                assert_eq!(ec.len(), 0);
        */
    }
}
